(*
The window description functions of the Window Contents Explorer utility
by Alexei Hmelnov.
----------------------------------------------------------------------------
E-Mail: alex@monster.icc.ru
http://monster.icc.ru/~alex/
----------------------------------------------------------------------------

See the file "readme.txt" for more details.

------------------------------------------------------------------------
                             IMPORTANT NOTE:
This software is provided 'as-is', without any expressed or implied warranty.
In no event will the author be held liable for any damages arising from the
use of this software.
Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:
1. The origin of this software must not be misrepresented, you must not
   claim that you wrote the original software.
2. Altered source versions must be plainly marked as such, and must not
   be misrepresented as being the original software.
3. This notice may not be removed or altered from any source
   distribution.
*)

unit WinDescr;

interface
uses
  Windows, SysUtils,CommCtrl, StlNames;

  function DescribeWin(H: HWND; P: TPoint; Ofs: integer): String;
  function DescrWindows(H: HWND; P: TPoint; Ofs: integer): String;
  function DescrWindowAtPos(P: TPoint): String;

const
  StyleNamesOn: boolean = true;

implementation

const
  SpaceS: ShortString = '';

  ExStyleNames: TCtlStyleRec = (
    Ranges: Nil;
    Names: (
    'WS_EX_DLGMODALFRAME' {1},
    '',
    'WS_EX_NOPARENTNOTIFY' {4},
    'WS_EX_TOPMOST' {8},
    'WS_EX_ACCEPTFILES' { $10},
    'WS_EX_TRANSPARENT' { $20},
    'WS_EX_MDICHILD' { $40},
    'WS_EX_TOOLWINDOW' { $80},
    'WS_EX_WINDOWEDGE' { $100},
    'WS_EX_CLIENTEDGE' { $200},
    'WS_EX_CONTEXTHELP' { $400},
    '',
    'WS_EX_RIGHT' { $1000},
    'WS_EX_RTLREADING' { $2000},
    'WS_EX_LEFTSCROLLBAR' { $4000},
    '');
  );

  ExStyleNames1: TStyleNames = (
    'WS_EX_CONTROLPARENT' { $10000},
    'WS_EX_STATICEDGE' { $20000},
    'WS_EX_APPWINDOW' { $40000},
    '','','','','','','','','','','','',''
  );

function StyleStr(S: DWORD; CtlSN: PCtlStyleRec; SNH: PStyleNames): String;
const
  WinStyleNames: TStyleNames = (
    'WS_TABSTOP' {= $10000},
    'WS_GROUP' {= $20000},
    'WS_THICKFRAME' {= $40000},
    'WS_SYSMENU' {= $80000},
    'WS_HSCROLL' {= $100000},
    'WS_VSCROLL' {= $200000},
    'WS_DLGFRAME' {= $400000},
    'WS_BORDER' {= $800000},
    'WS_MAXIMIZE' {= $1000000},
    'WS_CLIPCHILDREN' {= $2000000},
    'WS_CLIPSIBLINGS' {= $4000000},
    'WS_DISABLED' {= $8000000},
    'WS_VISIBLE' {= $10000000},
    'WS_MINIMIZE' {= $20000000},
    'WS_CHILD' {= $40000000},
    'WS_POPUP' {= $80000000}
  );

procedure PutFlag(SP: PChar);
begin
  if SP[0]<>'-' then begin
    if Result='' then
      Result := StrPas(SP)
    else
      Result := Format('%s|%s',[Result,SP]);
  end ;
end ;

var
  F,i,j: integer;
  SP: PChar;
  SN: PStyleNames;
  Range: PStyleRange;
begin
  {if S=0 then begin
    Result := '';
    Exit;
  end ;}
  Result := '';
  if StyleNamesOn then begin
    j := 0;
    SN := Nil;
    if CtlSN<>Nil then begin
      SN := @CtlSN^.Names;
      Range := CtlSN^.Ranges;
      while Range<>Nil do begin
        i := (S shr Range^.Base)and(Range^.Mask);
        if i<Range^.SzUsed then begin
          SP := Range^.Names^[i];
          if (SP<>Nil)and(SP[0]<>#0) then begin
            PutFlag(SP);
            S := S and not(Range^.Mask shl Range^.Base)
          end ;
        end ;
        Range := Range^.Next;
      end ;
    end ;
    if SNH=Nil then
      SNH := @WinStyleNames;
    repeat
      if SN=Nil then begin
        SN := SNH;
        j := 16;
      end ;
      for i:=0 to 15 do begin
        F := (1 shl j);
        SP := SN^[i];
        if (S and F<>0)and(SP<>Nil)and(SP[0]<>#0) then begin
          PutFlag(SP);
          S := S and not F;
        end ;
        Inc(j);
      end ;
      if SN=SNH then
        Break;
      SN := Nil;
    until false;
    if Result<>'' then begin
      if S=0 then
        Exit;
      Result := '|'+Result;
    end ;
  end ;
  Result := Format('0x%x%s',[S,result]);
end ;

function DescribeWin(H: HWND; P: TPoint; Ofs: integer): String;
var
  hItem,ID: integer;
  HSub: HWND;
  CNBuf,Txt:array[byte] of Char;
  Stl,ExtStl: LongInt;
  R: TRect;
  RS,StlS,ExtStlS: String;
  CtlSN: PCtlStyleRec;
  l: integer;

  procedure PutS(S: String);
  begin
    if l+Length(S)>80 then begin
      Result := Format('%s'#13#10'%s  ',[Result,SpaceS]);
      l := Ofs+2;
    end ;
    Result := Result+S;
    Inc(l,Length(S));
  end ;

begin
  hItem := GetWindowText(H,Txt,SizeOf(Txt)-1);
  hItem := GetClassName(H,CNBuf,SizeOf(CNBuf)-1);
  if GetWindowRect(H,R) then
    RS := Format('(%d,%d)[%d,%d]',[R.Left,R.Top,R.Right-R.Left,R.Bottom-R.Top])
  else
    RS := '()'{SetRectEmpty(R)};
  CtlSN := GetWinClassDescr(StrPas(CNBuf));
  Stl := GetWindowLong(H,GWL_STYLE);
  StlS := StyleStr(Stl,CtlSN,Nil);
  ExtStl := GetWindowLong(H,GWL_EXSTYLE);
  ExtStlS := StyleStr(ExtStl,@ExStyleNames,@ExStyleNames1);
  ID := GetDlgCtrlID(H);
  OffsetRect(R,-P.x,-P.y);
  SpaceS[0] := Chr(Ofs);
  FillChar(SpaceS[1],Ofs,' ');
  Result := Format('%s#%d:"%s":%s,R:%s,',
    [SpaceS,ID,Txt,CNBuf,RS]);
  PutS(Format('S:%s,',[StlS]));
  PutS(Format('ES:%s',[ExtStlS]));
end ;

var
  R: TRect;
function DescrWindows(H: HWND; P: TPoint; Ofs: integer): String;
var
  HSub: HWND;
begin
  if Ofs>255 then
    Ofs := 255;
  Result := DescribeWin(H,P,Ofs)+#13#10;
  if not GetWindowRect(H,R) then begin
    P.x := 0;
    P.y := 0;
   end
  else
    P := R.TopLeft;
  HSub := GetWindow(H,GW_CHILD);
  while HSub<>0 do begin
    Result := Result+DescrWindows(HSub,P,Ofs+2);
    HSub := GetWindow(HSub,GW_HWNDNEXT);
  end ;
end ;

function DescrWindowAtPos(P: TPoint): String;
var
  Wnd: HWND;
begin
  Wnd := WindowFromPoint(P);
  if Wnd=0 then
    Result := '?'
  else begin
    P.x := 0;
    P.y := 0;
    Result := DescrWindows(Wnd,P,0)
  end ;
end ;

end.
